//
//  Board.swift
//  GameBattleshipTerminal
//
//  Created by Piotr Fulmański on 2019.10.04.
//  Copyright © 2019 Piotr Fulmański. All rights reserved.
//

import Foundation

class Board {
    private var cols, rows: Int
    private var ships: Ships
    var shipsAtCommand: Int {
        return ships.shipsAtCommand
    }
    var board: [[CellContentType]]
    
    enum CellContentType {
        case none, empty, hit, notAllowed, rescue, ship, shot
    }
    
    init (rows: Int = 10, cols: Int = 10) {
        self.rows = rows
        self.cols = cols
        
        // Create a little bit bigger board with "safe area".
        board = Array(repeating:
                    Array(repeating:
                        CellContentType.none, count: cols+2),
                count: rows+2)
        
        // Fill board.
        for r in 0...rows+1 {
            for c in 0...cols+1 {
                board[r][c] = CellContentType.empty
            }
        }
        
        ships =  Ships()
    }
    
    private func afterHitAction(row: Int, col: Int) {
        for (_, ship) in ships.ships {
            if ship.isLocatedAt(row: row, col: col) {
                ship.hitAt(row: row, col: col)
                if ship.isDestroyed {
                    ship.autoMarkWhenDestroyed(onBoard: self)
                }
            }
        }
    }
    
    func mayPlaceShip(size: Int, anchor: (row: Int, col: Int), direction: Ship.Direction) -> Bool {
        var modifier: (forRow: Int, forCol: Int)!
        var actualRow, actualCol: Int!
        
        switch direction {
        case .up:
            modifier = (forRow: -1, forCol:  0)
        case .down:
            modifier = (forRow: +1, forCol:  0)
        case .left:
            modifier = (forRow:  0, forCol: -1)
        case .right:
            modifier = (forRow:  0, forCol: +1)
        }
        
        for i in 0...size-1 {
            actualRow = anchor.row + i*modifier.forRow
            actualCol = anchor.col + i*modifier.forCol
            
            guard (1...rows).contains(actualRow),
                (1...cols).contains(actualCol)
                else {
                    return false
            }
            
            if board[actualRow][actualCol] == CellContentType.ship
                || board[actualRow][actualCol] == CellContentType.notAllowed {
                return false
            }
        }
        
        return true
    }
    
    // First use "mayPlaceShip" method to be sure that ship place is possible.
    func placeShip(size: Int, anchor: (row: Int, col: Int), direction: Ship.Direction) {
        var modifier: (forRow: Int, forCol: Int)!
        var actualRow, actualCol: Int!
        var position = [(row: Int, col: Int, status: Ship.Status)]()
        
        switch direction {
        case Ship.Direction.up:
            modifier = (forRow: -1, forCol:  0)
        case .down:
            modifier = (forRow: +1, forCol:  0)
        case .left:
            modifier = (forRow:  0, forCol: -1)
        case .right:
            modifier = (forRow:  0, forCol: +1)
        }
        
        for i in 0...size-1 {
            actualRow = anchor.row + i*modifier.forRow
            actualCol = anchor.col + i*modifier.forCol
            
            position.append((row: actualRow, col: actualCol, status: Ship.Status.ready))
            board[actualRow][actualCol] = CellContentType.ship
            
            // To add border along ship
            actualRow = anchor.row+(modifier.forCol) + i*modifier.forRow
            actualCol = anchor.col+(modifier.forRow) + i*modifier.forCol
            board[actualRow][actualCol] = CellContentType.notAllowed
            
            actualRow = anchor.row-(modifier.forCol) + i*modifier.forRow
            actualCol = anchor.col-(modifier.forRow) + i*modifier.forCol
            board[actualRow][actualCol] = CellContentType.notAllowed
        }
        
        // To add borders at the ends of ship
        // Next to anchor
        actualRow = anchor.row + (-1)*modifier.forRow
        actualCol = anchor.col + (-1)*modifier.forCol
        board[actualRow][actualCol] = CellContentType.notAllowed
        
        actualRow = anchor.row+(modifier.forCol) + (-1)*modifier.forRow
        actualCol = anchor.col+(modifier.forRow) + (-1)*modifier.forCol
        board[actualRow][actualCol] = CellContentType.notAllowed
        
        actualRow = anchor.row-(modifier.forCol) + (-1)*modifier.forRow
        actualCol = anchor.col-(modifier.forRow) + (-1)*modifier.forCol
        board[actualRow][actualCol] = CellContentType.notAllowed
        
        // Next to anchor opposit end
        actualRow = anchor.row + (size)*modifier.forRow
        actualCol = anchor.col + (size)*modifier.forCol
        board[actualRow][actualCol] = CellContentType.notAllowed
        
        actualRow = anchor.row+(modifier.forCol) + (size)*modifier.forRow
        actualCol = anchor.col+(modifier.forRow) + (size)*modifier.forCol
        board[actualRow][actualCol] = CellContentType.notAllowed
        
        actualRow = anchor.row-(modifier.forCol) + (size)*modifier.forRow
        actualCol = anchor.col-(modifier.forRow) + (size)*modifier.forCol
        board[actualRow][actualCol] = CellContentType.notAllowed
        
        ships.ships["\(anchor)"] = Ship(size: size, position: position)
    }
    
    func mayShot(row: Int, col: Int) -> Bool {
        if board[row][col] == Board.CellContentType.empty ||
            board[row][col] == Board.CellContentType.ship ||
            board[row][col] == Board.CellContentType.notAllowed {
            return true
        } else if board[row][col] == Board.CellContentType.hit ||
            board[row][col] == Board.CellContentType.rescue ||
            board[row][col] == Board.CellContentType.ship ||
            board[row][col] == Board.CellContentType.shot {
            return false
        }
        
        return false
    }
    
    
    // Return the number of ships which were successfuly positioned on the board
    func shipsAutoSetup(shipsSize: [Int], maxTriesPerShip: Int) -> Int {
        var r, c: Int
        var shipDirection = Ship.Direction.up
        var possible = true
        var success: Bool
        var positioned = 0
        var anchor: (row: Int, col: Int)
        
        for size in shipsSize {
            success = false
            for _ in 1...maxTriesPerShip {
                r = Utils.getRandomInt(from: 1, to: rows)
                c = Utils.getRandomInt(from: 1, to: cols)
                anchor = (row: r, col: c)
                
                switch Utils.getRandomInt(from: 1, to: 4) {
                case 1: shipDirection = Ship.Direction.up
                case 2: shipDirection = Ship.Direction.right
                case 3: shipDirection = Ship.Direction.down
                default: shipDirection = Ship.Direction.left
                }
                possible = mayPlaceShip(size: size,
                                        anchor: anchor,
                                        direction: shipDirection)
                
                if possible {
                    placeShip(size: size,
                              anchor: anchor,
                              direction: shipDirection)
                    positioned += 1
                    success = true
                    break
                }
            }
            
            if !success {
                return positioned
            }
        }
        
        return positioned
    }
    
    func shot(row: Int, col: Int) {
        switch board[row][col] {
        case .empty:
            board[row][col] = CellContentType.shot
        case .ship:
            board[row][col] = CellContentType.hit
            afterHitAction(row: row, col: col)
        case .notAllowed:
            board[row][col] = CellContentType.shot
        case .hit:
            board[row][col] = CellContentType.hit
        case .rescue:
            board[row][col] = CellContentType.rescue
        case .none:
            board[row][col] = CellContentType.none
        case .shot:
            board[row][col] = CellContentType.shot
        }
    }
    
    // MARK: - DEBUG METHODS
    // This method is correct only for terminal application or debuging.
    // Consider move it to another class or extension.
    
    // Zostawić jako ćwiczenie
    // Etap 1: plansza z samymi kropkami
    // Etap 2: plansza z numerami na górze
    // Etap 3: plansza z numerami z lewej strony
    func printBoard() {
        let leadingPadding = Utils.determineNumberOfDigits(number: rows)
        var leadingPaddingString = ""
        var line = " "
        var digit: Int = 0
        
        for _ in 1...leadingPadding {
            leadingPaddingString += " "
        }
        
        line = " "+leadingPaddingString
        for c in 1...cols {
            digit = c/10
            if digit == 0 {
                line += " "
            } else {
                line += "\(digit)"
            }
        }
        print(line)
        
        line = " "+leadingPaddingString
        for c in 1...cols {
            digit = c%10
            line += "\(digit)"
        }
        print(line)
        
        line = leadingPaddingString
        for _ in 0...cols+1 {
            line += "="
        }
        print(line)
        
        for r in 0...rows+1 {
            line = ""
            
            if r == 0 || r == rows+1 {
                line = leadingPaddingString
            } else {
                line += String(r).leftPadding(toLength: leadingPadding)
            }
            
            for c in 0...cols+1 {
                switch board[r][c] {
                case CellContentType.empty:
                    line += "."
                case CellContentType.hit:
                    line += "!"
                case CellContentType.ship:
                    line += "X"
                case CellContentType.shot:
                    line += "*"
                case CellContentType.none:
                    line += "?"
                case CellContentType.notAllowed:
                    line += "+"
                case CellContentType.rescue:
                    line += "O"
                }
            }
            print(line)
        }
    }
}
