//
//  EngineGameBattleship.swift
//  GameBattleshipTerminal
//
//  Created by Piotr Fulmański on 2019.10.04.
//  Copyright © 2019 Piotr Fulmański. All rights reserved.
//

import Foundation

class EngineGameBattleship {
    private var boardPlayer: Board
    private var boardOpponent: Board
    // Cannot use Set, because Ship is not hashable.
    // Cannot use tuple as a hash in dictionary, because tuple is not hashable.
    // Can convert tuple into String and use this string as a key (String is hashable).
    // From:
    // https://stackoverflow.com/questions/24131323/in-swift-can-i-use-a-tuple-as-the-key-in-a-dictionary
    // To turn Ship into hashable use for example (to do as an excercise or in next iteration)
    // From:
    // https://www.embeddeduse.com/2015/12/30/swift-class-hashable/
    // https://useyourloaf.com/blog/swift-hashable/
    private var shipsSize: [Int]
    private var shipsSizePlayer: [Int]
    private var shipsSizeOpponent: [Int]
    private var rows, cols: Int
    
    
    enum Who {
        case player, opponent
    }
    
    init (rows: Int = 10, cols: Int = 10, shipsSize: [Int]) {
        boardPlayer = Board(rows: rows, cols: cols)
        boardOpponent = Board(rows: rows, cols: cols)
//        shipsPlayer = Ships()
//        shipsOpponent = Ships()
        
        self.shipsSize = shipsSize
        self.shipsSizePlayer = shipsSize
        self.shipsSizeOpponent = shipsSize
        
        self.rows = rows
        self.cols = cols
    }
    
    func checkWhoWins() -> Who? {
        if boardPlayer.shipsAtCommand == 0 {
            return Who.opponent
        } else if boardOpponent.shipsAtCommand == 0 {
            return Who.player
        }
        
        return nil
    }
    
    func getShootCoordinatesForOpponent(maxTries: Int) -> (row: Int, col: Int)? {
        var row, col: Int
        
        // Use random approach
        for _ in 1...maxTries {
            row = Utils.getRandomInt(from: 1, to: rows)
            col = Utils.getRandomInt(from: 1, to: cols)
            
            if mayShot(who: Who.opponent, row: row, col: col) {
                return (row: row, col: col)
            }
        }
        
        // If previous failed, use systematic search approach
        for r in 0...rows+1 {
            for c in 0...cols+1 {
                if mayShot(who: Who.opponent, row: r, col: c) {
                    return (row: r, col: c)
                }
            }
        }
        
        return nil
    }
    
    func mayPlaceShip(who: Who, size: Int, anchorRow: Int, anchorCol: Int, direction: Ship.Direction) -> Bool {
        let anchor = (row: anchorRow, col: anchorCol)
        let boardWho = getWhoBoard(who: who)
        return boardWho.mayPlaceShip(size: size, anchor: anchor, direction: direction)
    }
    
    // First use "mayPlaceShip" method to be sure that ship place is possible.
    func placeShip(who: Who, size: Int, anchorRow: Int, anchorCol: Int, direction: Ship.Direction) {
        let anchor = (row: anchorRow, col: anchorCol)
        let boardWho = getWhoBoard(who: who)
        boardWho.placeShip(size: size, anchor: anchor, direction: direction)
    }
    
    func mayShot(who: Who, row: Int, col: Int) -> Bool {
        let boardWho = getWhoBoard(who: getWhoTarget(who: who))
        return boardWho.mayShot(row: row, col: col)
    }
    
    // First use "mayShot" method to be sure that shoot is possible.
    func shot(who: Who, row: Int, col: Int) {
        let boardWho = getWhoBoard(who: getWhoTarget(who: who))
        boardWho.shot(row: row, col: col)
    }
    
    func shipsAutoSetup(shipsSize: [Int], maxTriesPerShip: Int, who: Who) -> Bool {
        let boardWho = getWhoBoard(who: who)
        let number = boardWho.shipsAutoSetup(shipsSize: shipsSize, maxTriesPerShip: maxTriesPerShip)
        
        if shipsSize.count == number {
            return true
        }
        
        return false
    }
    
    private func getWhoTarget(who: Who) -> Who {
        if who == Who.player {
            return Who.opponent
        }
        return Who.player
    }
    
    private func getWhoBoard(who: Who) -> Board {
        if who == Who.player {
            return boardPlayer
        }

        return boardOpponent
    }
    
    // MARK: - Debug and command line methods
    // BEGIN Debug and command line methods
    
    // From:
    // https://www.journaldev.com/19612/swift-readline-swift-print
    // W powyższym warto zwrócić uwagę na: Reading a 2D Array
    private func getIntFromCommandLine(message: String, rangeMin: Int, rangeMax: Int) -> Int {
        print(message)
        while(true) {
            if let input = readLine()
            {
                if let int = Int(input)
                {
                    if int >= rangeMin && int <= rangeMax {
                        return int
                    } else {
                        print("\(input) is not in range [\(rangeMin)-\(rangeMax)]. Please try againt")
                    }
                }
                else{
                    print("\(input) is not a valid integer. Please try againt")
                }
            }
        }
    }
    
    func getShootCoordinatesForPlayer() -> (row: Int, col: Int)? {
        while(true) {
            let row = getIntFromCommandLine(message: "Enter row", rangeMin: 1, rangeMax: rows)
            let col = getIntFromCommandLine(message: "Enter column", rangeMin: 1, rangeMax: cols)
            
            if mayShot(who: Who.player, row: row, col: col) {
                return (row: row, col: col)
            }
        }
    }
    
    func printBoards() {
        print("PLAYER")
        boardPlayer.printBoard()
        print("OPPONENT")
        boardOpponent.printBoard()
    }
    // END Debug and command line methods
}
